/***************************************************************************/
/* Copyright (C) 2006 Electronic Theatre Controls, Inc.                    */
/* All rights reserved.                                                    */
/*                                                                         */
/* This software is provided by Electronic Theatre Controls, Inc. ("ETC")  */
/* at no charge to the user to facilitate the use and implementation of    */
/* ANSI BSR E1.31-2006 BSR E1.31, Entertainment Technology - DMX512-A      */
/* Streaming Protocol (DSP).                                               */
/*                                                                         */
/* Redistribution and use in source and binary forms, with or without      */
/* modification, are permitted provided that the following conditions are  */
/* met:                                                                    */
/*                                                                         */
/*      Redistributions of source code, whether in binary form or          */
/*      otherwise must retain the above copyright notice, a list of        */
/*      modifications made, if any, a description of modifications made,   */
/*      this list of conditions and the following disclaimer.              */
/*                                                                         */
/* THIS SOFTWARE IS PROVIDED "AS IS" BY THE COPYRIGHT HOLDER AND WITHOUT   */
/* WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING, WITHOUT     */
/* LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,            */
/* MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.  THE     */
/* ENTIRE RISK AS TO THE QUALITY, OPERATION, AND PERFORMANCE OF THE        */
/* SOFTWARE IS WITH THE USER.                                              */
/*                                                                         */
/* UNDER NO CIRCUMSTANCES AND UNDER NO LEGAL THEORY, WHETHER TORT          */
/* (INCLUDING NEGLIGENCE), CONTRACT OR OTHERWISE, SHALL ETC BE LIABLE TO   */
/* ANY PERSON OR ENTITY FOR ANY INDIRECT, SPECIAL, INCIDENTAL, OR          */
/* CONSEQUENTIAL DAMAGES OF ANY KIND INCLUDING WITHOUT LIMITATION, DAMAGES */
/* FOR LOSS OF GOODWILL, WORK STOPPAGE, COMPUTER FAILURE OR MALFUNCTION,   */
/* OR ANY AND ALL OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF SUCH         */
/* PARTY SHALL HAVE BEEN INFORMED OF THE POSSIBILITY OF SUCH DAMAGES.      */
/* THIS LIMITATION OF LIABILITY SHALL NOT APPLY (I) TO DEATH OR PERSONAL   */
/* INJURY RESULTING FROM SUCH PARTY'S NEGLIGENCE TO THE EXTENT APPLICABLE  */
/* LAW PROHIBITS SUCH LIMITATION, (II) FOR CONSUMERS WITH RESIDENCE IN A   */
/* MEMBER COUNTRY OF THE EUROPEAN UNION (A) TO DEATH OR PERSONAL INJURY    */
/* RESULTING FROM ETC'S NEGLIGENCE, (B) TO THE VIOLATION OF ESSENTIAL      */
/* CONTRACTUAL DUTIES BY ETC, AND (C) TO DAMAGES RESULTING FROM            */
/* ETC'S GROSS NEGLIGENT OR WILLFUL BEHAVIOR, SOME JURISDICTIONS DO NOT    */
/* ALLOW THE EXCLUSION OR LIMITATION OF INCIDENTAL OR CONSEQUENTIAL        */
/* DAMAGES, SO THIS EXCLUSION AND LIMITATION MAY NOT APPLY TO YOU.         */
/*                                                                         */
/* Neither the name of ETC, its affiliates, nor the names of its           */
/* contributors may be used to endorse or promote products derived from    */
/* this software without express prior written permission of the President */
/* of ETC.                                                                 */
/*                                                                         */
/* The user of this software agrees to indemnify and hold ETC, its         */
/* affiliates, and contributors harmless from and against any and all      */
/* claims, damages, and suits that are in any way connected to or arising  */ 
/* out of, either in whole or in part, this software or the user's use of  */
/* this software.                                                          */
/*                                                                         */
/* Except for the limited license granted herein, all right, title and     */
/* interest to and in the Software remains forever with ETC.               */
/***************************************************************************/

//VHD.cpp		implementation of VHD functions

#include "deftypes.h"
#include "defpack.h"
#include "VHD.h"

//Defines for the VHD flags
const uint1 VHD_L_FLAG = 0x80;
const uint1 VHD_V_FLAG = 0x40;
const uint1 VHD_H_FLAG = 0x20;
const uint1 VHD_D_FLAG = 0x10;

//Given a pointer, packs the VHD inheritance flags.  The pointer returned is the 
//SAME pointer, since it is assumed that length will be packed next
uint1* VHD_PackFlags(uint1* pbuffer, bool inheritvec, bool inherithead, bool inheritdata)
{
	uint1* p = pbuffer;
	uint1 newbyte = UpackB1(p) & 0x8f;  //Mask out the VHD bits to keep the length intact

	//Update the byte flags
	if(!inheritvec)
		newbyte |= VHD_V_FLAG;
	if(!inherithead)
		newbyte |= VHD_H_FLAG;
	if(!inheritdata)
		newbyte |= VHD_D_FLAG;
	PackB1(p, newbyte);
	return pbuffer;
}

//Given a pointer, packs the length.  The pointer returned is the pointer
// to the rest of the buffer.  It is assumed that pbuffer contains at 
// least 3 bytes.  If inclength is true, the resultant length of the flags and
// length field is added to the length parameter before packing
uint1* VHD_PackLength(uint1* pbuffer, uint4 length, bool inclength)
{	
	uint1* p = pbuffer;
	uint4 mylen = length;
	if(inclength)
	{
		if(length + 1 > VHD_MAXMINLENGTH)
			mylen += 2;
		else
			++mylen;
	}

	uint1 newbyte = UpackB1(p) & 0x70;  //Mask out the length bits to keep the other flags intact
	//Set the length bit if necessary
	if(mylen > VHD_MAXMINLENGTH)
		newbyte |= VHD_L_FLAG;

	//pack the upper length bits
	uint1 packbuf[4];  //we have to manipulate the big endian bits
	PackB4(packbuf, mylen);
	if(mylen <= VHD_MAXMINLENGTH)
	{
		//Packbuf[0] and [1] should be empty, and the upper bits of [2] should be empty
		newbyte |= packbuf[2] & 0x0f;
		PackB1(p, newbyte);
		++p;
		PackB1(p, packbuf[3]);
		++p;
	}
	else
	{
		//Packbuf[0] and the upper bits of [1] are ignored
		//We give a max length constant, so the user is warned
		newbyte |= (packbuf[1] & 0x0f);
		PackB1(p, newbyte);
		++p;
		PackB1(p, packbuf[2]);
		++p;
		PackB1(p, packbuf[3]);
		++p;
	}
	return p;
}

//Given a pointer and vector size, packs the vector.  
// The pointer returned is the pointer to the rest of the buffer.
//It is assumed that pbuffer contains at least 4 bytes.
uint1* VHD_PackVector(uint1* pbuffer, uint4 vector, uint vecsize)
{
	uint1* p = pbuffer;
	if(vecsize == 1)
	{
		PackB1(pbuffer, static_cast<uint1>(vector));
		++p;
	}
	else if(vecsize == 2)
	{
		PackB2(pbuffer, static_cast<uint2>(vector));
		p += 2;
	}
	else
	{
		PackB4(pbuffer, vector);
		p += 4;
	}
	return p;
}

const uint1* VHD_GetFlagLength(const uint1* const pbuffer, bool& inheritvec,
									 bool& inherithead, bool& inheritdata, uint4& length)
{
	inheritvec = (*pbuffer & VHD_V_FLAG) == 0;
	inherithead = (*pbuffer & VHD_H_FLAG) == 0;
	inheritdata = (*pbuffer & VHD_D_FLAG) == 0;
	bool lensmall = (*pbuffer & VHD_L_FLAG) == 0;
	uint1 upackbuf[4];  //We need to manipulate bits
	if(lensmall)
	{
		upackbuf[0] = 0;
		upackbuf[1] = 0;
		upackbuf[2] = *pbuffer & 0x0f;
		upackbuf[3] = *(pbuffer + 1);
		length = UpackB4(upackbuf);
		return pbuffer + 2;	//flag/len byte and len byte
	}
	else
	{
		upackbuf[0] = 0;
		upackbuf[1] = *pbuffer & 0x0f;
		upackbuf[2] = *(pbuffer + 1);
		upackbuf[3] = *(pbuffer + 2);
		length = UpackB4(upackbuf);
		return pbuffer + 3; //flag/len byte, and 2 len bytes
	}
}


const uint1* VHD_GetVector1(const uint1* const pbuffer, uint1& vector)
{
	vector = UpackB1(pbuffer);
	return pbuffer + 1;
}

const uint1* VHD_GetVector2(const uint1* const pbuffer, uint2& vector)
{
	vector = UpackB2(pbuffer);
	return pbuffer + 2;
}

const uint1* VHD_GetVector4(const uint1* const pbuffer, uint4& vector)
{
	vector = UpackB4(pbuffer);
	return pbuffer + 4;
}
