/***************************************************************************/
/* Copyright (C) 2006 Electronic Theatre Controls, Inc.                    */
/* All rights reserved.                                                    */
/*                                                                         */
/* This software is provided by Electronic Theatre Controls, Inc. ("ETC")  */
/* at no charge to the user to facilitate the use and implementation of    */
/* ANSI BSR E1.31-2006 BSR E1.31, Entertainment Technology - DMX512-A      */
/* Streaming Protocol (DSP).                                               */
/*                                                                         */
/* Redistribution and use in source and binary forms, with or without      */
/* modification, are permitted provided that the following conditions are  */
/* met:                                                                    */
/*                                                                         */
/*      Redistributions of source code, whether in binary form or          */
/*      otherwise must retain the above copyright notice, a list of        */
/*      modifications made, if any, a description of modifications made,   */
/*      this list of conditions and the following disclaimer.              */
/*                                                                         */
/* THIS SOFTWARE IS PROVIDED "AS IS" BY THE COPYRIGHT HOLDER AND WITHOUT   */
/* WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING, WITHOUT     */
/* LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,            */
/* MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.  THE     */
/* ENTIRE RISK AS TO THE QUALITY, OPERATION, AND PERFORMANCE OF THE        */
/* SOFTWARE IS WITH THE USER.                                              */
/*                                                                         */
/* UNDER NO CIRCUMSTANCES AND UNDER NO LEGAL THEORY, WHETHER TORT          */
/* (INCLUDING NEGLIGENCE), CONTRACT OR OTHERWISE, SHALL ETC BE LIABLE TO   */
/* ANY PERSON OR ENTITY FOR ANY INDIRECT, SPECIAL, INCIDENTAL, OR          */
/* CONSEQUENTIAL DAMAGES OF ANY KIND INCLUDING WITHOUT LIMITATION, DAMAGES */
/* FOR LOSS OF GOODWILL, WORK STOPPAGE, COMPUTER FAILURE OR MALFUNCTION,   */
/* OR ANY AND ALL OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF SUCH         */
/* PARTY SHALL HAVE BEEN INFORMED OF THE POSSIBILITY OF SUCH DAMAGES.      */
/* THIS LIMITATION OF LIABILITY SHALL NOT APPLY (I) TO DEATH OR PERSONAL   */
/* INJURY RESULTING FROM SUCH PARTY'S NEGLIGENCE TO THE EXTENT APPLICABLE  */
/* LAW PROHIBITS SUCH LIMITATION, (II) FOR CONSUMERS WITH RESIDENCE IN A   */
/* MEMBER COUNTRY OF THE EUROPEAN UNION (A) TO DEATH OR PERSONAL INJURY    */
/* RESULTING FROM ETC'S NEGLIGENCE, (B) TO THE VIOLATION OF ESSENTIAL      */
/* CONTRACTUAL DUTIES BY ETC, AND (C) TO DAMAGES RESULTING FROM            */
/* ETC'S GROSS NEGLIGENT OR WILLFUL BEHAVIOR, SOME JURISDICTIONS DO NOT    */
/* ALLOW THE EXCLUSION OR LIMITATION OF INCIDENTAL OR CONSEQUENTIAL        */
/* DAMAGES, SO THIS EXCLUSION AND LIMITATION MAY NOT APPLY TO YOU.         */
/*                                                                         */
/* Neither the name of ETC, its affiliates, nor the names of its           */
/* contributors may be used to endorse or promote products derived from    */
/* this software without express prior written permission of the President */
/* of ETC.                                                                 */
/*                                                                         */
/* The user of this software agrees to indemnify and hold ETC, its         */
/* affiliates, and contributors harmless from and against any and all      */
/* claims, damages, and suits that are in any way connected to or arising  */ 
/* out of, either in whole or in part, this software or the user's use of  */
/* this software.                                                          */
/*                                                                         */
/* Except for the limited license granted herein, all right, title and     */
/* interest to and in the Software remains forever with ETC.               */
/***************************************************************************/

/*  tock.h

  Provides a standard definition of a tock and a ttimer.

  A tock is the number of milliseconds since a platform-specific date.
  Tocks are never used directly, rather the difference between two tocks 
  (latest - previous) are used to determine the passage of time.  It is
  assumed that tocks always move forward, and that the base date is not
  different for any instance of the compiled tock, nor does the base date change.  
  
  Tocks are guaranteed to have a millisecond granularity.
  Tock comparisons correctly handle clock rollover.

  A ttimer is a simple abstraction for typical timer usage, which is
  setting a number of milliseconds to time out, and then telling whether
  or not the ttimer has expired.
*/

#ifndef _TOCK_H_
#define _TOCK_H_

//tock requires the default types
#ifndef _DEFTYPES_H_
#error "#include error: tock.h requires deftypes.h"
#endif

class tock;
class ttimer;

//These functions comprise the tock API

//Initializes the tock layer.  Only needs to be called once per application,
//but can be called multiple times as long as there is an equal number of
//Tock_StopLib() calls.
bool Tock_StartLib();  
tock Tock_GetTock();   //Gets a tock representing the current time
void Tock_StopLib();   //Shuts down the tock layer.

//This is the actual tock 
class tock
{
public:
	//construction and copying
	tock();
	tock(uint4 ms);
	tock(const tock& t);
	tock& operator=(const tock& t);

	//Returns the number of milliseconds that this tock represents
	uint4 Getms();  

	//Used sparingly, but sets the number of milliseconds that this tock represents
	void Setms(uint4 ms);

protected:
	int4 v;  //Signed, so the wraparound calculations will work
	
	friend bool operator>(const tock& t1, const tock& t2);
	friend bool operator>=(const tock& t1, const tock& t2);
	friend bool operator==(const tock& t1, const tock& t2);
	friend bool operator!=(const tock& t1, const tock& t2);
	friend bool operator<(const tock& t1, const tock& t2);
	friend bool operator<=(const tock& t1, const tock& t2);
	friend uint4 operator-(const tock& t1, const tock& t2);
};

//The class used for simple expiration tracking
class ttimer
{
public:
	//construction/setup
	ttimer();				//Will immediately time out if timeout isn't set
	ttimer(int4 ms);	//The number of milliseconds before the timer will time out
	void SetInterval(int4 ms);	//Sets a new timeout interval (in ms) and resets the timer
	int4 GetInterval();			//Returns the current timeout interval (in ms)

	void Reset();	//Resets the timer, using the current timeout interval
	bool Expired();  //Returns true if the timer has expired.
					 //Call Reset() to use this timer again for a new interval.
protected:
	int4 interval;
	tock tockout;
};


//---------------------------------------------------------------------------------
//	Implementation

/*ttimer implementation*/
inline ttimer::ttimer():interval(0) {Reset();}
inline ttimer::ttimer(int4 ms):interval(ms) {Reset();}
inline void ttimer::SetInterval(int4 ms) {interval = ms; Reset();}
inline int4 ttimer::GetInterval() {return interval;}
inline void ttimer::Reset() {tockout.Setms(Tock_GetTock().Getms() + interval);}
inline bool ttimer::Expired() {return Tock_GetTock() > tockout;}

/*tock implementation*/
inline tock::tock():v(0) {}
inline tock::tock(uint4 ms):v(ms) {};
inline tock::tock(const tock& t) {v = t.v;}
inline tock& tock::operator=(const tock& t) {v = t.v; return *this;}

inline uint4 tock::Getms() {return v;}
inline void tock::Setms(uint4 ms) {v = ms;}
	
inline bool operator>(const tock& t1, const tock& t2)  {return t1.v - t2.v > 0;}
inline bool operator>=(const tock& t1, const tock& t2) {return t1.v - t2.v >= 0;}
inline bool operator==(const tock& t1, const tock& t2) {return t1.v - t2.v == 0;}
inline bool operator!=(const tock& t1, const tock& t2) {return t1.v - t2.v != 0;}
inline bool operator<(const tock& t1, const tock& t2)  {return t2.v - t1.v > 0;}
inline bool operator<=(const tock& t1, const tock& t2) {return t2.v - t1.v >= 0;}
inline uint4 operator-(const tock& t1, const tock& t2) {return t1.v - t2.v;}


#endif /*_TOCK_H*/
